#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "inttypes.h"

#include "disk.h"

#define FCB_FREE UINT32_MAX
#define FCB_USED (UINT32_MAX-1)

#define FAT_FREE UINT32_MAX
#define FAT_USED (UINT32_MAX-1)
#define FAT_EOF (UINT32_MAX-2)

#define DISK_FULL UINT32_MAX

typedef struct {
    char name[24];      // filnavn
    uint32_t length;    // fillngde
    uint32_t first;     // frste datablok
} fcb_t;

static uint32_t datasecs;      // antal datasektorer
static uint32_t *fat = NULL; // FAT buffer i hukommelse

static uint32_t fcbs         // antal FCB'er
    = SECTOR_SIZE/sizeof(fcb_t);
static fcb_t *dir;           // Katalog buffer i hukommelse

/* Allokerer en FCB og returnerer en pointer til den */
static fcb_t *alloc_fcb() {
    uint32_t i;

    for (i=0; i<fcbs; i++) {
        if (dir[i].first == FCB_FREE) {
            dir[i].first = FCB_USED;
            return dir+i;
        }
    }

    return NULL;
}

/* Allokerer en FCB og returnerer en pointer til den */
static uint32_t alloc_sector() {
    uint32_t i;

    for (i=0; i<datasecs; i++) {
        if (fat[i] == FAT_FREE) {
            fat[i] = FAT_USED;
            return i;
        }
    }

    return DISK_FULL;
}

/*
Aktiver filsystemet. Katalogsektor og FAT hentes ind i hukommelsen.
Ved fejl, returneres negativ fejlkode.
*/
int fs_mount() {
    uint32_t i;

    if (fat != NULL) {
        return -1; // allerede mount'et.
    }

    /* Antal datasektorer */
    datasecs = DISK_SECTORS*SECTOR_SIZE/(sizeof(unsigned int) + SECTOR_SIZE)-1;

    dir = malloc(SECTOR_SIZE);
    if (dir == NULL) {
        return -2; // ingen hukommelse;
    }

    fat = malloc(SECTOR_SIZE*(DISK_SECTORS - datasecs));
    if (fat == NULL) {
        free(dir);
        return -2; // ingen hukommelse;
    }

    /* ls katalogsektor */
    disk_read(datasecs, dir);

    /* ls FAT */
    for (i=datasecs+1; i<DISK_SECTORS; i++) {
        disk_read(i, ((void *)fat) + (i - datasecs - 1)*SECTOR_SIZE);
    }
    return 0;
}

/*
Deaktiver filsystemet. Katalogsektor og FAT
skrives ud p disk.
Ved fejl, returneres negativ fejlkode.
*/
int fs_unmount() {
    uint32_t i;

    if (fat == NULL) {
        return -1; // allerede unmount'et.
    }

    /* skriv katalogsektor */
    disk_write(datasecs, dir);

    /* skriv FAT */
    for (i=datasecs+1; i<DISK_SECTORS; i++) {
        disk_write(i, ((void *)fat) + (i - datasecs - 1)*SECTOR_SIZE);
    }

    free(fat);
    free(dir);
    fat = NULL;
    return 0;

}

/*
Initialiser et tomt filesystem p disken.
Ved fejl, returneres negativ fejlkode.
 */
int fs_format() {
    int e;
    uint32_t i;

    /*
    Hack: 'fs_mount' initialiserer de datastrukturer der skal bruges
    med hvad end der ligger p disken. De slettes dog nedefor.
    */
    e = fs_mount();
    if (e < 0) {
        return e;
    }

    /* Set alle 'first' til speciel vrdi som betyder at FCB'en er fri. */
    for (i=0; i<fcbs; i++) {
        dir[i].first = FCB_FREE;
    }

    /* Set alle links til speciel vrdi som betyder at blokken er fri. */
    for (i=0; i<datasecs; i++) {
        fat[i] = FAT_FREE;
    }

    /* Unmount for at skrive ndringer til disk */
    fs_unmount();
    return 0;
}

/*
Lister alle filenavne i filsystemet, samt disses lngde.
*/
int fs_ls() {
    uint32_t i;

    for (i=0; i<fcbs; i++) {
        if (dir[i].first != FCB_FREE) {
            printf("%s\t\t%u\n", dir[i].name, dir[i].length);
        }
    }
    return 0;
}

/*
Slet filen med navnet 'filename'.
Ved fejl, returner negativ fejlkode.
*/
int fs_delete(char *filename) {
	uint32_t i;
	uint32_t fcb_length;
    uint32_t fcb_first;
	uint32_t fat_next;
	int found = 0;
	
	/* Gennemlber kataloget ofr at finde filnavnet. Hvis det ikke findes, returneres -1, 
	   ellers gemmes length og first, hvorefter disse og filnavnet nulstilles: */
	
	for (i=0; i<fcbs; i++) {
        if (strcmp(dir[i].name, filename) == 0) {
			found = 1;
			break;
        }
    }

	if(!found) {
		printf("FEJL (fs_delete): Fil ikke fundet!\n");
		return -1; //file not found
	}
	
	fcb_length = dir[i].length;
	fcb_first = dir[i].first;
	strcpy(dir[i].name, "");
	dir[i].length = 0;
	dir[i].first = FCB_FREE;

	/* Vi flger en eventuel hgtet liste i FAT'en, hvor vi stter alle filens sektorer til FAT_FREE: */
	
	fat_next = fcb_first;
	uint32_t temp;
	while(fat_next != FAT_EOF) {
		temp = fat_next;
		fat_next = fat[fat_next];
		fat[temp] = FAT_FREE;
	}	
	fat[fat_next] = FAT_FREE;
	
	printf("Slettet fil: %s\n", filename);
	
    return 0;
}

/*
Ls hele filens indhold ind i 'data'.
Ved fejl, returner negativ fejlkode.
Det antages at filen kan vre i 'data'.
*/
int fs_read(char *filename, void *data) {
	uint32_t i;
	uint32_t fcb_length;
    uint32_t fcb_first;
	int found = 0;
	
	/* Gennemlber kataloget ofr at finde filnavnet. Hvis det ikke findes, returneres -1, 
	   ellers ...: */
	
	for (i=0; i<fcbs; i++) {
        if (strcmp(dir[i].name, filename) == 0) {
			found = 1;
			printf("Filnavn fundet i katalog\n");
			break;
        }
    }

	if(!found) {
		printf("FEJL (fs_read): Fil ikke fundet!\n");
		return -1; //File not found
	}

	fcb_length = dir[i].length;
	i = fcb_first = dir[i].first;

	/* Lser sektorer pnr den sidste. Koden er nrmest en kopi af fs_write(): */
	while (fcb_length >= SECTOR_SIZE) {
		disk_read(i, data);
		fcb_length -= SECTOR_SIZE;
		data += SECTOR_SIZE;
		if (fcb_length > 0) {
			i = fat[i];
		}
	}

	/* Lser sidste sektor. Her skal vi kun bruge resten af filen (dvs. mske ikke 
	hele den sidste sektor). Derfor lser vi hele sektoren fra disken og kopierer 
	den relevante del ind i data med memcpy: */
	if (fcb_length > 0) {
		uint8_t sector[SECTOR_SIZE];
		disk_read(i, sector);
		memcpy(data, sector, fcb_length);
	}
	printf("S er vi frdige med fs_read\n");
	return 0;
}

/*
Skriver 'len' bytes fra 'data' til en fil med navnet 'filename'.
Hvis filen allerede eksisterer, s overskrives den gamle.
Ved fejl, returneres negativ fejlkode.

For at gre koden let tilgngelig hndteres fejl ikke!
*/
int fs_write(char *filename, void *data, uint32_t len) {
    uint32_t i;
    fcb_t *fcb;

    /* Slet evt. eksisterende fil med samme navn */
    fs_delete(filename);

    fcb = alloc_fcb();

    /* Initialiser FCB */
    strcpy(fcb->name, filename);
    fcb->length = len;
    i = fcb->first = alloc_sector();

    /* Gem sektorer */
    while (len >= SECTOR_SIZE) {
        disk_write(i, data);
        len -= SECTOR_SIZE;
        data += SECTOR_SIZE;
        if (len > 0) {
            fat[i] = alloc_sector();
            i = fat[i];
        }
    }

    /* Gem sidste sektor */
    if (len > 0) {
        uint8_t sector[SECTOR_SIZE];
        memcpy(sector, data, len);
        disk_write(i, sector);
    }
    fat[i] = FAT_EOF;

    return 0;
}

int main() {
    char *str="This is a testy test string";
	char test[30];

    fs_format();

    fs_mount();

    fs_write("test1", str, strlen(str)+1);
    fs_write("test2", str, 20);
    fs_write("test3", str, 10);

    fs_unmount();

    fs_mount();

    fs_ls();

	fs_delete("test2\0");
	fs_ls();
	fs_delete("test2\0");
	fs_ls();

	fs_read("test3\0", &test);
	printf("Data udtrukket fra test3: %s\n", test);
	
    fs_unmount();

    return 0;
}
